/*
 xppmview: inclemental ppm viewer

 (c) 1998 Yutaka Oiwa.

*/
#ifdef DEBUG
define dprintf fprintf
#else
inline void dprintf(...){}
#endif

#if defined(sun) && !defined(__svr4__)
#define __USE_FIXED_PROTOTYPES__
extern "C" {
  int ioctl(int, int, char *);
}
#endif

static char rcsid[] = "$Id: xppmview.cc,v 1.8 1999/01/13 15:33:57 yutaka Exp $";

#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xresource.h>

#include <X11/Xaw/Command.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Box.h>

#if defined(sun) && defined(__svr4__)
#include <sys/filio.h>
#endif


// Dither Parameters
int R_DEPTH, G_DEPTH, B_DEPTH, R_STEP, G_STEP, B_STEP;
int R_SHIFT, G_SHIFT, B_SHIFT;

const int diff_shift[11][10] = 
{{-1},
 {0}, 
 {0,1},
 {0,1,0},
 {0,1,0,3},
 {0,2,0,2,0},
 {0,3,0,3,0,3},
 {0,4,0,2,4,0,4},
 {0,5,0,3,0,5,0,3},
 {0,5,0,4,0,3,0,5,0},
 {0,7,0,4,0,6,0,3,0,5}};

#define D_SIZE 4
const int  dith_order[D_SIZE][D_SIZE] = {{ 1,  9,  3, 11},
					 {13,  5, 15,  7},
					 { 4, 12,  2, 10},
					 {16,  8, 14,  6}};

int  dith_red[D_SIZE][D_SIZE];
int  dith_blue[D_SIZE][D_SIZE];
int  dith_green[D_SIZE][D_SIZE];

void initialize_dither(int table[D_SIZE][D_SIZE], int depth, int *step)
{
  int x, y;
  *step = 256 / depth;
//  printf("depth %d -> step %d\n", depth, *step);
  for(y = 0; y < D_SIZE; y++)
    for(x = 0; x < D_SIZE; x++)
      table[y][x] = (dith_order[y][x] * 2 + 1) * *step / (D_SIZE * D_SIZE * 2);
}

/*
   Widget Tree:

   ToplevelShell shell
     +- Paned paned
         +- Core canvas
	 +- Box btn_bar
	     +- Command quit_btn
*/

XtAppContext acon = NULL;
Widget shell, paned, canvas, btn_bar, quit_btn, message;
Display *disp;
Window win;
Visual *visual;
int visual_class;
int visual_depth;
int bytes_per_pixel;
Screen *scr;
GC gc;
char *image_buffer = NULL;
XImage *ximage;
bool is_file;
bool widget_initialized = false;
Pixel bp, wp;

int ifd = 0, ofd = 0;
XtInputId input_id = 0;
XtBlockHookId block_id = 0;
XtIntervalId timer_id = 0;
Colormap    cmap;
typedef unsigned long p_table_t[8][16][8];
p_table_t palette_table;

enum pnm_type_t { PPM, PGM, PBM };
pnm_type_t pnm_type;
bool pnm_raw_p;
unsigned int pnm_width, pnm_height;
unsigned int canvas_width, canvas_height;
unsigned int interior_width, interior_height;
unsigned int image_bpl;

struct app_data_t {
  Bool help;
  int magnify;
  Bool diffusion_p;
  char *if_name;
  char *of_name;
} app_data;

enum state_t { READ_SIGNATURE, READ_SIZE, READ_DATA, BAD_EOF, PREMATURE_EOF,
	       EOF_REACHED, READ_DEAD, FINISHED };
state_t state;

unsigned int cur_x = 0, cur_y = 0;
unsigned int repaint_start_x = 0;
unsigned int repaint_start_y = 0;

enum read_type_t { READ_LINE, READ_WORD, READ_BYTE };
read_type_t read_type = READ_LINE;
void (*arrive_proc)(char *, int) = NULL;

void quit_proc(Widget, XtPointer, XtPointer)
{
  exit(0);
}
void quit_event(Widget, XEvent *, String *, Cardinal *)
{
  exit(0);
}

void show_message(char *fmt, ...)
{
  if (!widget_initialized) return;
  
  va_list ap;
  va_start(ap, fmt);
  char buf[512];
  vsprintf(buf, fmt, ap);
  va_end(ap);
  
  XtVaSetValues(message, XtNlabel, buf, NULL);
}

void select_visual()
{
  int nitems;
  XVisualInfo *info;
  XVisualInfo temp;
  
  temp.screen = DefaultScreen(disp);
  info = XGetVisualInfo(disp, VisualScreenMask, &temp, &nitems);
  int preffered = -1;
  int flag = 0;
  int curdepth = 0;
  for(int i = 0; i < nitems; i++) {
//    printf("visual %d: id %ld depth %d class %d\n",
//	   i, info[i].visualid, info[i].depth, info[i].c_class);
    if(info[i].c_class == TrueColor) {
      if (flag < 2 || info[i].depth >= curdepth) {
	if (info[i].depth < 15) continue;
	preffered = i;
	curdepth = info[i].depth;
	visual_class = info[i].c_class;
	visual = info[i].visual;
	visual_depth = info[i].depth;
	flag = 2;
      }
    }
    if(info[i].c_class == PseudoColor) {
      if (flag < 1 || info[i].depth >= curdepth) {
	if (info[i].depth != 8) continue;
	preffered = i;
	curdepth = info[i].depth;
	visual_class = info[i].c_class;
	visual = info[i].visual;
	visual_depth = info[i].depth;
	flag = 1;
      }
    }
  }
//  printf ("preffered = %d\n", preffered);
  if (info) XFree((XPointer)info);
  if (!flag) {
    fprintf(stderr, "Visual not Available.\n");
    exit(1);
  }
}
  
// lacks prototype
extern "C" void XtResizeWidget(Widget w, 
			       Dimension width, 
			       Dimension height, 
			       Dimension border_width);

void initialize_widget(unsigned w, unsigned h)
{
  if (! widget_initialized) {
    widget_initialized = true;
    if (!w) {
      w = 128, h = 128;
    }
    // Search for Visual
    paned = XtVaCreateManagedWidget("paned", panedWidgetClass, shell, NULL);
    canvas = XtVaCreateManagedWidget("canvas", simpleWidgetClass, paned, 
				     XtNwidth, w,
				     XtNheight, h,
				     NULL);
    message = XtVaCreateManagedWidget("message", labelWidgetClass, paned, 
				      XtNwidth, 192, 
				      XtNlabel, "",
				      NULL);
    btn_bar = XtVaCreateManagedWidget("box", boxWidgetClass, paned, NULL);
    quit_btn = XtVaCreateManagedWidget("quit", commandWidgetClass, btn_bar, NULL);
    XtAddCallback(quit_btn, XtNcallback, quit_proc, NULL);
    select_visual();
    XtVaSetValues(canvas, XtNvisual, visual, NULL);

    XtRealizeWidget(shell);
    
    win = XtWindow(canvas);
    gc = XCreateGC(disp, win, 0, 0);
    
    unsigned short width, height;
    XtVaGetValues(shell, 
		  XtNwidth, &width,
		  XtNheight, &height,
		  NULL);
    interior_height = height - h;
    interior_width = width;
  } else {
    if (!w) return;
    Position width = w > interior_width ? w : interior_width;
    Position height = h + interior_height;

    XtResizeWidget(shell, width, height, 0);
    XtVaSetValues(shell, 
/*		  XtNwidth, width,
		  XtNheight, height,*/
		  XtNmaxWidth, width,
		  XtNmaxHeight, height,
		  XtNminWidth, width,
		  XtNminHeight, height,
		  NULL);
    XtResizeWidget(canvas, w, h, 0);
    XtVaSetValues(canvas,
/*		  XtNwidth, w,
		  XtNheight, h,*/
		  XtNmin, h,
		  XtNmax, h,
		  NULL);
  }
}

void dispatch_data(int c)
{
  static char lbuf[512];
  static unsigned int lbuf_len = 0;
  
  switch(read_type) {
  case READ_BYTE:
    arrive_proc(NULL, c);
    return;
  case READ_WORD:
    if (c == EOF || isspace(c)) {
      if (c == EOF || lbuf_len > 0) {
	lbuf[lbuf_len] = 0;
	arrive_proc(lbuf, c);
	lbuf_len = 0;
      }
      return;
    }
    break;
  case READ_LINE:
    if (c == EOF || c == '\n') {
      lbuf[lbuf_len] = 0;
      arrive_proc(lbuf, c);
      lbuf_len = 0;
      return;
    }
    break;
  }
  if (lbuf_len + 2 >= sizeof lbuf) {
    if (lbuf_len + 2 == sizeof lbuf)
      fprintf(stderr, "data_arrived: too long input %s.\n", read_type == READ_WORD? "word" : "line");
    lbuf_len++;
  } else
    lbuf[lbuf_len++] = c;
}

void data_arrived_callback(XtPointer, int *, XtInputId *)
{
  static char buf[512];
  int r = read(ifd, buf, 512);
  if (r < 0) {
    perror("read_byte(read)");
  }
  if (r == 0)
    dispatch_data(EOF);
  else {
    if (ofd) {
      if (write(ofd, buf, r) != r) {
	perror("read_byte(thru_out)");
      }
    }
    for(int i = 0; i < r; i++)
      dispatch_data((unsigned char)buf[i]);
  }
}

// forward declaration

void set_state(state_t new_state);

void read_signature(char *p, int) 
{
  if (strcmp(p, "P6") == 0) {
    pnm_type = PPM;
    pnm_raw_p = true;
    set_state(READ_SIZE);
    show_message("PPM(B) signature arrived.");
    return;
  }
  if (strcmp(p, "P3") == 0) {
    pnm_type = PPM;
    pnm_raw_p = false;
    set_state(READ_SIZE);
    show_message("PPM(A) signature arrived.");
    return;
  }
  initialize_widget(0, 0);
  show_message("bad PPM signature.");
  set_state(READ_DEAD);
}

enum { SIZE_X, SIZE_Y, SIZE_MAX } size_state = SIZE_X;

void read_size(char *p, int c) {
  char *endptr;
  int i = strtol(p, &endptr, 10);
  if (!*p || *endptr) {
    initialize_widget(0, 0);
    show_message("Bad size specification.");
    set_state(READ_DEAD);
    return;
  }
  switch(size_state) {
  case SIZE_X:
    pnm_width = i;
    size_state = SIZE_Y;
    break;
  case SIZE_Y:
    pnm_height = i;
    size_state = SIZE_MAX;
    break;
  case SIZE_MAX:
    if (c != '\n') {
      initialize_widget(0, 0);
      show_message("read_size: bad size specification.");
      set_state(READ_DEAD);
      return;
    }
    show_message("size: %d x %d (%d).", pnm_width, pnm_height, i);
    set_state(READ_DATA);
    break;
  }
}

/* Create color map */
int create_colormap(Display *d, Colormap *cmap, p_table_t cmap2)
{
  int err = 0;
  int blue, red, green;
  XColor col;

  for (blue = 0; blue <= B_DEPTH; blue++) {
    for (red = 0; red <= R_DEPTH; red++) {
      for (green = 0; green <= G_DEPTH; green++) {
	col.red = red * R_STEP;
	if (col.red > 255)
	  col.red = 255;
	col.red *= 257;
	
	col.green = green * G_STEP;
	if (col.green > 255)
	  col.green = 255;
	col.green *= 257;
	
	col.blue = blue * B_STEP;
	if (col.blue > 255)
	  col.blue = 255;
	col.blue *= 257;
	
	if (!XAllocColor(d, *cmap, &col)) {
	  if (!err) {
	    err = 1;
	    *cmap = XCopyColormapAndFree(d, *cmap);
	    if (*cmap == 0)
	      return 2;
	    if (!XAllocColor(d, *cmap, &col)) {
	      return 2;
	    }
	  } else {
	    return 2;
	  }
	}
	/* printf("col.pixel:%08x \n",col.pixel); */
	wp = cmap2[red][green][blue] = col.pixel;
      }
    }
  }
  bp = cmap2[0][0][0];
  return err;
}

void repaint_hook(XtPointer);

void (*put_pixel_image)(int, int, int, int, int) = NULL;
void put_pixel_image_palette(int x, int y, int r, int g, int b)
{
  image_buffer[y * canvas_width + x] = palette_table[r][g][b];
}

void put_pixel_image_direct(int x, int y, int r, int g, int b)
{
  XPutPixel(ximage, x, y, r << R_SHIFT | g << G_SHIFT | b << B_SHIFT);
}

void calc_shift(unsigned int mask, int *shift, int *depth)
{
  int i = 0;
  while(mask && (mask & 1) == 0) {
    mask>>=1, i++;
  }
  if(!mask) abort();
  if(mask & (mask + 1)) abort(); // must be 2^n - 1 for some n
  *shift = i;
  *depth = mask;
//  printf("shift %d depth %d\n", *shift, *depth);
}

void put_pixel(int cred, int cgreen, int cblue)
{
  static int R_diff = 0, G_diff = 0, B_diff = 0;
  if (cur_x == 0 &&  cur_y == 0) {
    R_diff = R_STEP / 2;
    G_diff = G_STEP / 2;
    B_diff = B_STEP / 2;
  }
  int x = cur_x * app_data.magnify;
  int y = cur_y * app_data.magnify;
  if (app_data.diffusion_p) {		/* Another dithering mode */
    for (int i = 0; i < app_data.magnify; i++) {
      for (int j = 0; j < app_data.magnify; j++) {
	/* difference difussion */
	R_diff += cred;
	int r_c = R_diff * R_DEPTH / 255;
	if (r_c > R_DEPTH) r_c = R_DEPTH;
	if (r_c < 0) r_c = 0;
	R_diff -= r_c * 255 / R_DEPTH;

	G_diff += cgreen;
	int g_c = G_diff * G_DEPTH / 255;
	if (g_c > G_DEPTH) g_c = G_DEPTH;
	if (g_c < 0) g_c = 0;
	G_diff -= g_c * 255 / G_DEPTH;

	B_diff += cblue;
	int b_c = B_diff * B_DEPTH / 255;
	if (b_c > B_DEPTH) b_c = B_DEPTH;
	if (b_c < 0) b_c = 0;
	B_diff -= b_c * 255 / B_DEPTH;

/*
	int     r_c = cred / R_STEP;
	int     g_c = cgreen / G_STEP;
	int     b_c = cblue / B_STEP;
	
	int r_real = r_c * R_STEP;
	int g_real = g_c * G_STEP;
	int b_real = b_c * B_STEP;
	
	int     rd = cred - r_real + R_diff;
	int     gd = cgreen - g_real + G_diff;
	int     bd = cblue - b_real + B_diff;
	if ( rd > R_STEP * DITH_FACTOR) r_c++;
	if ( gd > G_STEP * DITH_FACTOR) g_c++;
	if ( bd > B_STEP * DITH_FACTOR) b_c++;
	if (-rd > R_STEP * DITH_FACTOR) r_c--;
	if (-gd > G_STEP * DITH_FACTOR) g_c--;
	if (-bd > B_STEP * DITH_FACTOR) b_c--;
	if (r_c > R_DEPTH)	  r_c = R_DEPTH - 1;
	if (g_c > G_DEPTH)	  g_c = G_DEPTH - 1;
	if (b_c > B_DEPTH)	  b_c = B_DEPTH - 1;
	if (r_c < 0)	  r_c = 0;
	if (g_c < 0)	  g_c = 0;
	if (b_c < 0)	  b_c = 0;
	    
	r_real = r_c * R_STEP;
	g_real = g_c * G_STEP;
	b_real = b_c * B_STEP;

	R_diff -= r_real - cred;
	G_diff -= g_real - cgreen;
	B_diff -= b_real - cblue;
*/
	put_pixel_image(x + i, 
			y + (j + diff_shift[app_data.magnify][i]) % app_data.magnify,
			r_c, g_c, b_c);
      }
    }
  } else {			/* Original dithering */
    int red2, red3, res_r;
    int green2, green3, res_g;
    int blue2, blue3, res_b;
    
    res_r = cred % R_STEP;
    red3 = cred / R_STEP;
    res_g = cgreen % G_STEP;
    green3 = cgreen / G_STEP;
    res_b = cblue % B_STEP;
    blue3 = cblue / B_STEP;
    
    for (int j = 0; j < app_data.magnify; j++) {
      for (int i = 0; i < app_data.magnify; i++) {
	int dith_x = (x + i) % D_SIZE;
	int dith_y = (y + j) % D_SIZE;
	
	red2 = red3 + (res_r >= dith_red[dith_y][dith_x]);
	if (red2 > R_DEPTH) red2 = R_DEPTH;
	green2 = green3 + (res_g >= dith_green[dith_y][dith_x]);
	if (green2 > G_DEPTH) green2 = G_DEPTH;
	blue2 = blue3 + (res_b >= dith_blue[dith_y][dith_x]);
	if (blue2 > B_DEPTH) blue2 = B_DEPTH;
	
	put_pixel_image(x + i, y + j, red2, green2, blue2);
      }
    }
  }

  // Send Update Event
  if (!block_id && !is_file)
    block_id = XtAppAddBlockHook(acon, repaint_hook, NULL);
  
  if (++cur_x >= pnm_width) {
    cur_x = 0;
    if (++cur_y >= pnm_height) {
      set_state(FINISHED);
    }
    if (cur_y % (128 / app_data.magnify)  == 0)
      repaint_hook((XtPointer)True);
  }
}

void eof_check(char *, int c)
{
  static int num = 0;
  if (c == EOF) {
    set_state(EOF_REACHED);
    return;
  }
  if (!num) {
    show_message("warning: Too Many Data. Ignored.");
  }
  num++;
}

void trash_data(char *, int c) {
  if(c == EOF) {
    set_state(BAD_EOF);
    return;
  }
}

void read_data(char *p, int c)
{
  static int read_buf[3];
  static int s = 0;
  if (c == EOF) {
    show_message("Early EOF Detected @ [%d,%d].", cur_x, cur_y);
    set_state(PREMATURE_EOF);
  }
  if (pnm_raw_p)
    read_buf[s++] = c;
  else
    read_buf[s++] = atoi(p);
  if(s == 3) {
    put_pixel(read_buf[0], read_buf[1], read_buf[2]);
    s = 0;
  }
}

void set_shell_size(int width_, int height_)
{
  int w = width_, h = height_;

  XtVaSetValues(shell,
		XtNwidth, w,
		XtNheight, h,
		XtNwidth, w,
		XtNheight, h,
		XtNwidth, w,
		XtNheight, h,
		NULL);
}

void timeout_callback(XtPointer, XtIntervalId *)
{
  static unsigned int px = (unsigned)-1, py = 0;
  timer_id = XtAppAddTimeOut(acon, 1000, timeout_callback, NULL);
  if (state == READ_DATA) {
    if (cur_x == px && cur_y == py)
      return;
    px = cur_x;
    py = cur_y;
    show_message("[%d,%d] (%d%%)", cur_x, cur_y,
		 (cur_y * pnm_width + cur_x) * 100 / (pnm_height * pnm_width));
  }
}

void set_state_read_data() {
  canvas_width = pnm_width * app_data.magnify;
  canvas_height = pnm_height * app_data.magnify;
  cur_x = cur_y = 0;
  
  initialize_widget(canvas_width, canvas_height);

  if (visual_class == PseudoColor) {
//    printf("Creating Colormap.\n");
    R_DEPTH = 4;
    G_DEPTH = 8;
    B_DEPTH = 3;
    initialize_dither(dith_red, R_DEPTH, &R_STEP);
    initialize_dither(dith_green, G_DEPTH, &G_STEP);
    initialize_dither(dith_blue, B_DEPTH, &B_STEP);

    XtVaGetValues(shell, XtNcolormap, &cmap, NULL);
    switch (create_colormap(disp, &cmap, palette_table)) {
    case 0:
      break;
    case 1:
      XtVaSetValues(shell, XtNcolormap, cmap, NULL);
      break;
    case 2:
      break;
    }
    put_pixel_image = put_pixel_image_palette;
    bytes_per_pixel = 1;
    image_buffer = new char [canvas_width * canvas_height];
    ximage = XCreateImage( disp, visual, 8, ZPixmap, 
			  0, image_buffer, 
			  canvas_width, canvas_height, 8, 0);
    for(unsigned y = 0; y < canvas_height; y++)
      for(unsigned x = 0; x < canvas_width; x++)
	image_buffer[y * canvas_width + x] =
	  (((y + x) & 3) && ((y - x) & 3) ? bp : wp);
  } else {
    calc_shift(visual->red_mask,   &R_SHIFT, &R_DEPTH);
    calc_shift(visual->green_mask, &G_SHIFT, &G_DEPTH);
    calc_shift(visual->blue_mask,  &B_SHIFT, &B_DEPTH);
    initialize_dither(dith_red, R_DEPTH, &R_STEP);
    initialize_dither(dith_green, G_DEPTH, &G_STEP);
    initialize_dither(dith_blue, B_DEPTH, &B_STEP);

    bytes_per_pixel = (visual_depth + 7) / 8;
    int pad = bytes_per_pixel * 8;
    if (pad <= 8) pad = 8;
    else if (pad <= 16) pad = 16;
    else pad = 32;

    image_buffer = new char [canvas_width * canvas_height * pad / 8];
    ximage = XCreateImage (disp, visual, visual_depth, ZPixmap,
			   0, image_buffer, 
			   canvas_width, canvas_height, pad, 0);
//    printf("%d %d\n", canvas_width, canvas_height);
//    printf("%d %d %d %d\n", ximage->width, ximage->height, 
//	   ximage->bytes_per_line, ximage->depth);
    put_pixel_image = put_pixel_image_direct;
    for(unsigned y = 0; y < canvas_height; y++)
      for(unsigned x = 0; x < canvas_width; x++)
	if (((y+x) & 3) && ((y - x) & 3))
	  put_pixel_image(x, y, R_DEPTH, G_DEPTH, B_DEPTH);
	else
	  put_pixel_image(x, y, 0, 0, 0);
  }
  if(!ximage) {
    fprintf(stderr, "cannot get XImage.\n");
    exit(1);
  }
  XPutImage(disp, win, gc, ximage, 0, 0, 0, 0, canvas_width, canvas_height);

  char buf[512], *old_title;
  XtVaGetValues(shell,
		XtNtitle, &old_title, 
		NULL);

  char *p = strrchr(app_data.if_name, '/');
  if (p)
    p++;
  else
    p = app_data.if_name;
  sprintf(buf, "%.200s: %.200s (%c %dx%d)",
	  old_title, p, 'A' + pnm_raw_p, pnm_width, pnm_height);
  XtVaSetValues(shell, 
		XtNtitle, buf,
		NULL);

  timer_id = XtAppAddTimeOut(acon, 1000, timeout_callback, NULL);

  read_type = pnm_raw_p ? READ_BYTE : READ_WORD;
  arrive_proc = read_data;
}

void set_state(state_t new_state) {
  switch(state) {
  case READ_DATA:
    if (timer_id)
      XtRemoveTimeOut(timer_id);
    timer_id = 0;
    break;
  default:;
  }
  switch(new_state) {
  case READ_SIGNATURE:
    read_type = READ_LINE;
    arrive_proc = read_signature;
    break;
  case READ_SIZE:
    read_type = READ_WORD;
    arrive_proc = read_size;
    break;
  case READ_DATA:
    set_state_read_data();
    break;
  case READ_DEAD:
    read_type = READ_BYTE;
    arrive_proc = trash_data;
    break;
  case FINISHED:
    repaint_hook((XtPointer)2);
    show_message("Finished.");
    read_type = READ_BYTE;
    arrive_proc = eof_check;
    break;
  case BAD_EOF:
  case PREMATURE_EOF:
  case EOF_REACHED:
    if (input_id)
      XtRemoveInput(input_id);
    repaint_hook((XtPointer)2);
    input_id = 0;
    break;
  }
  state = new_state;
}

// repaint if blocked
void repaint_hook(XtPointer p)
{
  block_id = 0;
  if (p == NULL) {
    int n;
    ioctl(ifd, FIONREAD, (char *)&n);
    if (n > 0) {
      return;
    }
  }

  unsigned x, y, width, height;

  y = repaint_start_y;
  if (repaint_start_y == cur_y) {
    height = 1;
    // in one line
    if (repaint_start_x == cur_x) 
      return;
    width = cur_x - repaint_start_x;
    x = repaint_start_x;
  } else {
    height = cur_y - repaint_start_y + (cur_x != 0);
    x = 0;
    width = pnm_width;
  }

  dprintf(stderr, "repaint (%d,%d)->(%d,%d), (%d,%d)[%d,%d]\n",
	  repaint_start_x, repaint_start_y,
	  cur_x, cur_y,
	  x, y,
	  width, height);

  y *= app_data.magnify;
  x *= app_data.magnify;
  width *= app_data.magnify;
  height *= app_data.magnify;
    
  XPutImage(disp, win, gc, ximage, x, y, 
	    x, y, width, height);

  repaint_start_x = cur_x;
  repaint_start_y = cur_y;

  return;
}

static void expose_event(Widget, XEvent *e, String *, Cardinal *)
{
  static bool first = 1;
  if (first) {
    first = 0;
    if (state != FINISHED && state != EOF_REACHED 
	&& state != PREMATURE_EOF) {
      // First Expose Event is derivered almost asynchronously.
      // if still reading data, turned into repaint event in future.
      repaint_start_x = 0;
      repaint_start_y = 0;
      return;
    }
  } if (!e) {
    dprintf(stderr, "event ptr is NULL.\n");
    return;
  }
  if (e->type != Expose) {
    dprintf(stderr, "not a expose event.\n");
    return;
  }
  if (!image_buffer) {
    return;
  }
  XExposeEvent *p = &(e->xexpose);
  dprintf(stderr, "Expose Event (%d,%d)[%d,%d]<%d>[%d]\n",
	  p->x, p->y, p->width, p->height, p->send_event, p->count);
  XPutImage(disp, win, gc, ximage, p->x, p->y, 
	    p->x, p->y, p->width, p->height);
}

static XtActionsRec actions[] = {
  {"expose", expose_event}, 
  {"quit", quit_event}, 
};

String default_resources[] = {
  "*paned*showGrip: false", 
  "*quit.label: Quit",
  "*message.justify: left",
  "*translations: #override <Key>q: quit()",
  "*canvas.translations: #override <Expose>: expose()",
  NULL,
};

XrmOptionDescRec options[] = {
  { "-h", ".help", XrmoptionNoArg, (XPointer)"True" },
  { "-m", ".magnify", XrmoptionSepArg, (XPointer)NULL },
  { "-d", ".diffusion", XrmoptionNoArg, (XPointer)"True" },
  { "-i", ".inputFile", XrmoptionSepArg, (XPointer)NULL },
  { "-o", ".outputFile", XrmoptionSepArg, (XPointer)NULL },
};

XtResource resources[] = {
#define offset(ent) XtOffset(struct app_data_t *, ent)
  { "help", "Help", XtRBoolean, sizeof(Boolean),
    offset(help), XtRImmediate, (XtPointer)False }, 
  { "magnify", "Magnify", XtRInt, sizeof(int),
    offset(magnify), XtRImmediate, (XtPointer)2 },
  { "diffusion", "Diffusion", XtRBoolean, sizeof(Boolean),
    offset(diffusion_p), XtRImmediate, (XtPointer)False },
  { "inputFile", "InputFile", XtRString, sizeof(String),
    offset(if_name), XtRImmediate, (XtPointer)NULL },
  { "outputFile", "outputFile", XtRString, sizeof(String),
    offset(of_name), XtRImmediate, (XtPointer)NULL },
#undef offset
};

int main(int argc, char **argv)
{
  shell = XtVaAppInitialize
    (&acon, "XPpmView", options, XtNumber(options),
     &argc, argv, default_resources, NULL);
  if (!shell) {
    fprintf(stderr, "can't open display\n");
    return 1;
  }
  disp = XtDisplay(shell);

  XtAppAddActions(acon, actions, XtNumber(actions));
  XtGetApplicationResources(shell, XtPointer(&app_data), resources,
			    XtNumber(resources), NULL, 0);

  if (!app_data.if_name && argc >= 2 && (argv[1][0] != '-')) {
    app_data.if_name = argv[1];
    argc--; argv++;
  }
  if (argc >= 2 && strcmp(argv[1], "-") == 0) {
    argc--; argv++;
  }
  if (argc >= 2 
      || app_data.magnify < 1 || app_data.magnify > 10
      || app_data.help) {
    fprintf(stderr, "%s\n", rcsid);
    fprintf(stderr, "Usage: xppmview [-d] [-m magnify(1-10)]\n");
    fprintf(stderr, "                [-o outfile] [[-i] inputfile]\n");
    return 0;
  }

  if (app_data.if_name) {
    if (strcmp(app_data.if_name, "-") == 0) {
      // No Operation
    } else {
      ifd = open(app_data.if_name, O_RDONLY);
      if (ifd < 0) {
	perror("main(read_open)");
	exit(1);
      }
    }
  } else {
    app_data.if_name = "-";
  }
  if (app_data.of_name) {
    if (strcmp(app_data.of_name, "-") == 0) {
      ofd = 1;
    } else {
      ofd = open(app_data.of_name, O_WRONLY | O_CREAT | O_TRUNC, 0666);
      if (ofd < 0) {
	perror("main(write_open)");
	exit(1);
      }
    }
  }
  struct stat sbuf;
  if (fstat(ifd, &sbuf) < 0)
    is_file = false;
  else
    is_file = S_ISREG(sbuf.st_mode);

  input_id = XtAppAddInput(acon, ifd, (XtPointer) XtInputReadMask,
			   data_arrived_callback, NULL);

  if (!is_file)
    initialize_widget(0, 0);

  show_message("Waiting for data...");
  set_state(READ_SIGNATURE);
  
  XtAppMainLoop(acon);
}
